#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "process_path.h"
#include "construct_path.h"
#include "path_parser.tab.h"

/* Flattens a list of lists into a single list */
command_list_t* flatten(command_list_list_t* path) {
  command_t* first = path->head->head;
  command_list_t* cur_list = path->head;
  free(path);
  while(cur_list->next) {
    command_t* cur = cur_list->head;
    while(cur->next) {
      cur = cur->next;
    }
    command_list_t* old_list = cur_list;
    cur_list = cur_list->next;
    cur->next = cur_list->head;
    free(old_list);
  }
  cur_list->head = first;
  return cur_list;
}

/* Walks a list of commands and applies the following changes:
 * - transforms all commands into absolute coordinates
 * - populates the x0 and y0 start point fields for each command
 * - transforms hline and vline commands into equivalent line commands
 * - transforms smooth bezier curves into their non shorthand equivalents
 * - transforms close path commands into their equivalent line commands
 */
void normalize(command_list_t* path) {
  // flags
  char last_segment_closed_path = 0;
  char last_segment_was_cubic_curve = 0;
  char last_segment_was_quad_curve = 0;
  float last_control_point_x, last_control_point_y;

  command_t* cur = path->head;

  // always starts with an absolute move command
  float x = cur->x;
  float y = cur->y;
  float subpath_start_x = x;
  float subpath_start_y = y;
  cur->abs_or_rel = ABSOLUTE;
  cur->x0 = 0.0;
  cur->y0 = 0.0;
  cur = cur->next;

  // walk the list of commands
  while(cur) {
    // set start point fields
    cur->x0 = x;
    cur->y0 = y;
    switch(cur->type) {
      case MOVE_CMD:
        switch(cur->abs_or_rel) {
          case RELATIVE:
            cur->x = cur->x + x;
            cur->y = cur->y + y;
            cur->abs_or_rel = ABSOLUTE;
          case ABSOLUTE:
            x = cur->x;
            y = cur->y;
            if (last_segment_closed_path) {
              subpath_start_x = x;
              subpath_start_y = y;
            }
        }
        last_segment_closed_path = 0;
        last_segment_was_cubic_curve = 0;
        last_segment_was_quad_curve = 0;
        break;
      case LINE_CMD:
        switch(cur->abs_or_rel) {
          case RELATIVE:
            cur->x = cur->x + x;
            cur->y = cur->y + y;
            cur->abs_or_rel = ABSOLUTE;
          case ABSOLUTE:
            x = cur->x;
            y = cur->y;
        }
        last_segment_closed_path = 0;
        last_segment_was_cubic_curve = 0;
        last_segment_was_quad_curve = 0;
        break;
      case HLINE_CMD:
        cur->type = LINE_CMD;
        cur->y = y;
        switch(cur->abs_or_rel) {
          case RELATIVE:
            cur->x = cur->x + x;
            cur->abs_or_rel = ABSOLUTE;
          case ABSOLUTE:
            x = cur->x;
        }
        last_segment_closed_path = 0;
        last_segment_was_cubic_curve = 0;
        last_segment_was_quad_curve = 0;
        break;
      case VLINE_CMD:
        cur->type = LINE_CMD;
        cur->x = x;
        switch(cur->abs_or_rel) {
          case RELATIVE:
            cur->y = cur->y + y;
            cur->abs_or_rel = ABSOLUTE;
          case ABSOLUTE:
            y = cur->y;
        }
        last_segment_closed_path = 0;
        last_segment_was_cubic_curve = 0;
        last_segment_was_quad_curve = 0;
        break;
      case CUBIC_CURVE_CMD:
        switch(cur->abs_or_rel) {
          case RELATIVE:
            cur->control_point_data.x1 = cur->control_point_data.x1 + x;
            cur->control_point_data.y1 = cur->control_point_data.y1 + y;
            cur->control_point_data.x2 = cur->control_point_data.x2 + x;
            cur->control_point_data.y2 = cur->control_point_data.y2 + y;
            cur->x  = cur->x + x;
            cur->y  = cur->y + y;
            cur->abs_or_rel = ABSOLUTE;
          case ABSOLUTE:
            x = cur->x;
            y = cur->y;
        }
        last_control_point_x = cur->control_point_data.x2;
        last_control_point_y = cur->control_point_data.y2;
        last_segment_closed_path = 0;
        last_segment_was_cubic_curve = 1;
        last_segment_was_quad_curve = 0;
        break;
      case SMOOTH_CUBIC_CURVE_CMD:
        switch(cur->abs_or_rel) {
          case RELATIVE:
            cur->control_point_data.x2 = cur->control_point_data.x2 + x;
            cur->control_point_data.y2 = cur->control_point_data.y2 + y;
            cur->x  = cur->x + x;
            cur->y  = cur->y + y;
            cur->abs_or_rel = ABSOLUTE;
          case ABSOLUTE:
            cur->type = CUBIC_CURVE_CMD;
            if (last_segment_was_cubic_curve) {
              // reflection of last control point about current point
              cur->control_point_data.x1 = 2*x - last_control_point_x;
              cur->control_point_data.y1 = 2*y - last_control_point_y;
            } else {
              // coincident with current point
              cur->control_point_data.x1 = x;
              cur->control_point_data.y1 = y;
            }
            x = cur->x;
            y = cur->y;
        }
        last_control_point_x = cur->control_point_data.x2;
        last_control_point_y = cur->control_point_data.y2;
        last_segment_closed_path = 0;
        last_segment_was_cubic_curve = 1;
        last_segment_was_quad_curve = 0;
        break;
      case QUAD_CURVE_CMD:
        switch(cur->abs_or_rel) {
          case RELATIVE:
            cur->control_point_data.x1 = cur->control_point_data.x1 + x;
            cur->control_point_data.y1 = cur->control_point_data.y1 + y;
            cur->x  = cur->x + x;
            cur->y  = cur->y + y;
            cur->abs_or_rel = ABSOLUTE;
          case ABSOLUTE:
            x = cur->x;
            y = cur->y;
        }
        last_control_point_x = cur->control_point_data.x1;
        last_control_point_y = cur->control_point_data.y1;
        last_segment_closed_path = 0;
        last_segment_was_cubic_curve = 0;
        last_segment_was_quad_curve = 1;
        break;
      case SMOOTH_QUAD_CURVE_CMD:
        switch(cur->abs_or_rel) {
          case RELATIVE:
            cur->x  = cur->x + x;
            cur->y  = cur->y + y;
            cur->abs_or_rel = ABSOLUTE;
          case ABSOLUTE:
            cur->type = QUAD_CURVE_CMD;
            if (last_segment_was_quad_curve) {
              // reflection of last control point about current point
              cur->control_point_data.x1 = 2*x - last_control_point_x;
              cur->control_point_data.y1 = 2*y - last_control_point_y;
            } else {
              // coincident with current point
              cur->control_point_data.x1 = x;
              cur->control_point_data.y1 = y;
            }
            x = cur->x;
            y = cur->y;
        }
        last_control_point_x = cur->control_point_data.x1;
        last_control_point_y = cur->control_point_data.y1;
        last_segment_closed_path = 0;
        last_segment_was_cubic_curve = 0;
        last_segment_was_quad_curve = 1;
        break;
      case ARC_CMD:
        switch(cur->abs_or_rel) {
          case RELATIVE:
            cur->x = cur->x + x;
            cur->y = cur->y + y;
            cur->abs_or_rel = ABSOLUTE;
          case ABSOLUTE:
            x = cur->x;
            y = cur->y;
        }
        last_segment_closed_path = 0;
        last_segment_was_cubic_curve = 0;
        last_segment_was_quad_curve = 0;
        break;
      case CLOSE_PATH_CMD:
        cur->abs_or_rel = ABSOLUTE;
        cur->x = subpath_start_x;
        cur->y = subpath_start_y;
        cur->type = LINE_CMD;
        x = subpath_start_x;
        y = subpath_start_y;
        last_segment_closed_path = 1;
        last_segment_was_cubic_curve = 0;
        last_segment_was_quad_curve = 0;
        break;
    }
    cur = cur->next;
  }
}

/* Walks a list of commands and applies the following changes:
 * - scales all coordinates from the space [0,original_size] to the space [offset,255-offset]
 * - makes the path begin and end with moves from and to the center
 */
void scale_and_close(command_list_t* path, float original_size, float offset) {
  float size = 255.0 - 2*offset;
  float scale_factor = size/original_size;
	
  command_t *cur, *prev;
  cur = path->head;
  float center = 255.0/2.0;

  // always starts with a move from the center
  cur->x0 = center;
  cur->y0 = center;
  cur->x = (cur->x)*scale_factor+offset;
  cur->y = (cur->y)*scale_factor+offset;
  cur = cur->next;

  // walk the list of commands
  while(cur) {
    cur->x0 = (cur->x0)*scale_factor+offset;
    cur->y0 = (cur->y0)*scale_factor+offset;
    if (cur->type == ARC_CMD) {
      cur->arc_data.rx = (cur->arc_data.rx)*scale_factor+offset;
      cur->arc_data.ry = (cur->arc_data.ry)*scale_factor+offset;
    } else {
      cur->control_point_data.x1 = (cur->control_point_data.x1)*scale_factor+offset;
      cur->control_point_data.y1 = (cur->control_point_data.y1)*scale_factor+offset;
      cur->control_point_data.x2 = (cur->control_point_data.x2)*scale_factor+offset;
      cur->control_point_data.y2 = (cur->control_point_data.y2)*scale_factor+offset;
    }
    cur->x = (cur->x)*scale_factor+offset;
    cur->y = (cur->y)*scale_factor+offset;
    prev = cur;
    cur = cur->next;
  }

  // append a move to the center to the end
  float x = prev->x;
  float y = prev->y;
  command_t* closing_move = new_move_command(ABSOLUTE, center, center);
  closing_move->x0 = x;
  closing_move->y0 = y;
  concat(path, closing_move);
}

void rotate(command_list_t* path, float cx, float cy, float theta) {
  command_t* cur = path->head;
  float sin_theta = sin(theta);
  float cos_theta = cos(theta);
  while(cur) {
    float new_x0 = cx + cos_theta*(cur->x0 - cx) - sin_theta*(cur->y0 - cy);
    float new_y0 = cy + sin_theta*(cur->x0 - cx) + cos_theta*(cur->y0 - cy);
    float new_x1 = cx + cos_theta*(cur->control_point_data.x1 - cx) - sin_theta*(cur->control_point_data.y1 - cy);
    float new_y1 = cy + sin_theta*(cur->control_point_data.x1 - cx) + cos_theta*(cur->control_point_data.y1 - cy);
    float new_x2 = cx + cos_theta*(cur->control_point_data.x2 - cx) - sin_theta*(cur->control_point_data.y2 - cy);
    float new_y2 = cy + sin_theta*(cur->control_point_data.x2 - cx) + cos_theta*(cur->control_point_data.y2 - cy);
    float new_x  = cx + cos_theta*(cur->x - cx) - sin_theta*(cur->y - cy);
    float new_y  = cy + sin_theta*(cur->x - cx) + cos_theta*(cur->y - cy);

    cur->x0 = new_x0;
    cur->y0 = new_y0;
    cur->control_point_data.x1 = new_x1;
    cur->control_point_data.y1 = new_y1;
    cur->control_point_data.x2 = new_x2;
    cur->control_point_data.y2 = new_y2;
    cur->x = new_x;
    cur->y = new_y;

    cur = cur->next;
  }
}

/* Converts a command to its character encoding */
char cmd_to_char(command_t* cmd) {
  const char* cmd_letters = "MmLlHhVvCcSsQqTtAaZz";
  return cmd_letters[(int)(cmd->type)*2 + (int)(cmd->abs_or_rel)];
}

/* Prints a list of commands in svg path format to a file */
void print_path_svg(FILE* output, command_list_t* path) {
  command_t* cur = path->head;
  int large_arc_flag, sweep_flag;
  while(cur) {
    char cmd_char = cmd_to_char(cur);
    switch(cur->type) {
      case MOVE_CMD:
      case LINE_CMD:
        fprintf(output, "%c %f %f\n",
                cmd_char,
                cur->x,
                cur->y);
        break;
      case HLINE_CMD:
        fprintf(output, "%c %f\n",
                cmd_char,
                cur->x);
        break;
      case VLINE_CMD:
        fprintf(output, "%c %f\n",
                cmd_char,
                cur->y);
        break;
      case CUBIC_CURVE_CMD:
        fprintf(output, "%c %f %f %f %f %f %f\n",
                cmd_char,
                cur->control_point_data.x1,
                cur->control_point_data.y1,
                cur->control_point_data.x2,
                cur->control_point_data.y2,
                cur->x,
                cur->y);
        break;
      case SMOOTH_CUBIC_CURVE_CMD:
        fprintf(output, "%c %f %f %f %f\n",
                cmd_char,
                cur->control_point_data.x2,
                cur->control_point_data.y2,
                cur->x,
                cur->y);
        break;
      case QUAD_CURVE_CMD:
        fprintf(output, "%c %f %f %f %f\n",
                cmd_char,
                cur->control_point_data.x1,
                cur->control_point_data.y1,
                cur->x,
                cur->y);
        break;
      case SMOOTH_QUAD_CURVE_CMD:
        fprintf(output, "%c %f %f\n",
                cmd_char,
                cur->x,
                cur->y);
        break;
      case ARC_CMD:
        large_arc_flag = cur->arc_data.flags & 1;
        sweep_flag = (cur->arc_data.flags & (1 << 1)) >> 1;
        fprintf(output, "%c %f %f %f %d %d %f %f\n",
                cmd_char,
                cur->arc_data.rx,
                cur->arc_data.ry,
                cur->arc_data.x_axis_rotation,
                large_arc_flag,
                sweep_flag,
                cur->x,
                cur->y
              );
        break;
      case CLOSE_PATH_CMD:
        fprintf(output, "%c\n", cmd_char);
    }
    cur = cur->next;
  }
}

/* Prints a list of commands in longhand format to a file */
void print_path_full(FILE* output, command_list_t* path) {
  command_t* cur = path->head;
  int large_arc_flag, sweep_flag;
  while(cur) {
    char cmd_char = cmd_to_char(cur);
    switch(cur->type) {
      case MOVE_CMD:
      case LINE_CMD:
        fprintf(output,
          "%c\n\tX0:%10.6f Y0:%10.6f\n\tX1:%10.6f Y1:%10.6f\n",
                cmd_char,
                cur->x0,
                cur->y0,
                cur->x,
                cur->y);
        break;
      case HLINE_CMD:
        fprintf(output,
          "%c\n\tX0:%10.6f\n\tX1:%10.6f\n",
                cmd_char,
                cur->x0,
                cur->x);
        break;
      case VLINE_CMD:
        fprintf(output,
          "%c\n\tY0:%10.6f\n\tY1:%10.6f\n",
                cmd_char,
                cur->y0,
                cur->y);
        break;
      case CUBIC_CURVE_CMD:
        fprintf(output,
          "%c\n\tX0:%10.6f Y0:%10.6f\n\tX1:%10.6f Y1:%10.6f\n\tX2:%10.6f Y2:%10.6f\n\tX3:%10.6f Y3:%10.6f\n",
                cmd_char,
                cur->x0,
                cur->y0,
                cur->control_point_data.x1,
                cur->control_point_data.y1,
                cur->control_point_data.x2,
                cur->control_point_data.y2,
                cur->x,
                cur->y);
        break;
      case SMOOTH_CUBIC_CURVE_CMD:
        fprintf(output,
          "%c\n\tX0:%10.6f Y0:%10.6f\n\tX2:%10.6f Y2:%10.6f\n\tX3:%10.6f Y3:%10.6f\n",
                cmd_char,
                cur->x0,
                cur->y0,
                cur->control_point_data.x2,
                cur->control_point_data.y2,
                cur->x,
                cur->y);
        break;
      case QUAD_CURVE_CMD:
        fprintf(output,
          "%c\n\tX0:%10.6f Y0:%10.6f\n\tX1:%10.6f Y1:%10.6f\n\tX2:%10.6f Y2:%10.6f\n",
                cmd_char,
                cur->x0,
                cur->y0,
                cur->control_point_data.x1,
                cur->control_point_data.y1,
                cur->x,
                cur->y);
        break;
      case SMOOTH_QUAD_CURVE_CMD:
        fprintf(output,
          "%c\n\tX0:%10.6f Y0:%10.6f\n\tX2:%10.6f Y2:%10.6f\n",
                cmd_char,
                cur->x0,
                cur->y0,
                cur->x,
                cur->y);
        break;
      case ARC_CMD:
        large_arc_flag = cur->arc_data.flags & 1;
        sweep_flag = (cur->arc_data.flags & (1 << 1)) >> 1;
        fprintf(output,
          "%c\n\tX0:%10.6f Y0:%10.6f\n\tRX:%10.6f RY:%10.6f\n\tROT:%10.6f\n\tA:%d S:%d\n\tX1:%10.6f Y1:%10.6f\n",
                cmd_char,
                cur->x0,
                cur->y0,
                cur->arc_data.rx,
                cur->arc_data.ry,
                cur->arc_data.x_axis_rotation,
                large_arc_flag,
                sweep_flag,
                cur->x,
                cur->y);
        break;
      case CLOSE_PATH_CMD:
        fprintf(output, "%c\n", cmd_char);
    }
    cur = cur->next;
  }
}

/* Prints a list of commands in longhand format to a file */
void print_path_hdl(FILE* output, command_list_t* path) {
  int n = list_length(path);
  int type[n], x0[n], x1[n], x2[n], x3[n], y0[n], y1[n], y2[n], y3[n];
  command_t* cur = path->head;
  int i = 0;
  while(cur) {
    switch(cur->type) {
      case MOVE_CMD:
        type[i] = SOLVER_MOVE;
        x0[i] = float2fix(cur->x0);
        x1[i] = float2fix(cur->x);
        x2[i] = 0;
        x3[i] = 0;
        y0[i] = float2fix(cur->y0);
        y1[i] = float2fix(cur->y);
        y2[i] = 0;
        y3[i] = 0;
        break;
      case LINE_CMD:
        type[i] = SOLVER_LINE;
        x0[i] = float2fix(cur->x0);
        x1[i] = float2fix(cur->x);
        x2[i] = 0;
        x3[i] = 0;
        y0[i] = float2fix(cur->y0);
        y1[i] = float2fix(cur->y);
        y2[i] = 0;
        y3[i] = 0;
        break;
      case CUBIC_CURVE_CMD:
        type[i] = SOLVER_CUBIC;
        x0[i] = float2fix(cur->x0);
        x1[i] = float2fix(cur->control_point_data.x1);
        x2[i] = float2fix(cur->control_point_data.x2);
        x3[i] = float2fix(cur->x);
        y0[i] = float2fix(cur->y0);
        y1[i] = float2fix(cur->control_point_data.y1);
        y2[i] = float2fix(cur->control_point_data.y2);
        y3[i] = float2fix(cur->y);
        break;
      case QUAD_CURVE_CMD:
        type[i] = SOLVER_CUBIC;
        x0[i] = float2fix(cur->x0);
        x1[i] = float2fix(cur->control_point_data.x1);
        x2[i] = float2fix(cur->x);
        x3[i] = 0;
        y0[i] = float2fix(cur->y0);
        y1[i] = float2fix(cur->control_point_data.y1);
        y2[i] = float2fix(cur->y);
        y3[i] = 0;
        break;
      default:
        fprintf(stderr, "Unexpected %c command after normalization\n", cmd_to_char(cur));
        exit(0);
        break;
    }
    i++;
    cur = cur->next;
  }
  fprintf(output, "type_mem[0:%d] = \'{", n);
  for(i = 0; i < n; i++) {
    fprintf(output, "3\'d%x, ", type[i]);
  }
  fprintf(output, "3\'d4};\n");

  fprintf(output, "x0_mem[0:%d] = \'{", n-1);
  for(i = 0; i < n-1; i++) {
    fprintf(output, "28\'sh%07x, ", x0[i]);
  }
  fprintf(output, "28\'sh%07x};\n", x0[i]);

  fprintf(output, "x1_mem[0:%d] = \'{", n-1);
  for(i = 0; i < n-1; i++) {
    fprintf(output, "28\'sh%07x, ", x1[i]);
  }
  fprintf(output, "28\'sh%07x};\n", x1[i]);

  fprintf(output, "x2_mem[0:%d] = \'{", n-1);
  for(i = 0; i < n-1; i++) {
    fprintf(output, "28\'sh%07x, ", x2[i]);
  }
  fprintf(output, "28\'sh%07x};\n", x2[i]);

  fprintf(output, "x3_mem[0:%d] = \'{", n-1);
  for(i = 0; i < n-1; i++) {
    fprintf(output, "28\'sh%07x, ", x3[i]);
  }
  fprintf(output, "28\'sh%07x};\n", x3[i]);

  fprintf(output, "y0_mem[0:%d] = \'{", n-1);
  for(i = 0; i < n-1; i++) {
    fprintf(output, "28\'sh%07x, ", y0[i]);
  }
  fprintf(output, "28\'sh%07x};\n", y0[i]);

  fprintf(output, "y1_mem[0:%d] = \'{", n-1);
  for(i = 0; i < n-1; i++) {
    fprintf(output, "28\'sh%07x, ", y1[i]);
  }
  fprintf(output, "28\'sh%07x};\n", y1[i]);

  fprintf(output, "y2_mem[0:%d] = \'{", n-1);
  for(i = 0; i < n-1; i++) {
    fprintf(output, "28\'sh%07x, ", y2[i]);
  }
  fprintf(output, "28\'sh%07x};\n", y2[i]);

  fprintf(output, "y3_mem[0:%d] = \'{", n-1);
  for(i = 0; i < n-1; i++) {
    fprintf(output, "28\'sh%07x, ", y3[i]);
  }
  fprintf(output, "28\'sh%07x};\n\n", y3[i]);

  fprintf(output, "type_mem[%d:127] = \'{%d{3'b0}};\n",n+1, 127-n);
  fprintf(output, "x0_mem[%d:127] = \'{%d{28'b0}};\n",n, 127-n+1);
  fprintf(output, "x1_mem[%d:127] = \'{%d{28'b0}};\n",n, 127-n+1);
  fprintf(output, "x2_mem[%d:127] = \'{%d{28'b0}};\n",n, 127-n+1);
  fprintf(output, "x3_mem[%d:127] = \'{%d{28'b0}};\n",n, 127-n+1);
  fprintf(output, "y0_mem[%d:127] = \'{%d{28'b0}};\n",n, 127-n+1);
  fprintf(output, "y1_mem[%d:127] = \'{%d{28'b0}};\n",n, 127-n+1);
  fprintf(output, "y2_mem[%d:127] = \'{%d{28'b0}};\n",n, 127-n+1);
  fprintf(output, "y3_mem[%d:127] = \'{%d{28'b0}};\n",n, 127-n+1);
}
